import logging

from O365.connection import Connection
from O365.fluent_message import Message

log = logging.getLogger(__name__)


class FluentInbox(object):
	url_dict = {
		'inbox': {
			'1.0': 'https://outlook.office365.com/api/v1.0/me/messages',
			'2.0': 'https://graph.microsoft.com/v1.0/me/messages',
		},
		'folders': {
			'1.0': 'https://outlook.office365.com/api/v1.0/me/Folders',
			'2.0': 'https://graph.microsoft.com/v1.0/me/MailFolders',
		},
		'folder': {
			'1.0': 'https://outlook.office365.com/api/v1.0/me/Folders/{folder_id}/messages',
			'2.0': 'https://graph.microsoft.com/v1.0/me/MailFolders/{folder_id}/messages',
		},
		'child_folders': {
			'1.0': 'https://outlook.office365.com/api/v1.0/me/Folders/{folder_id}/childfolders',
			'2.0': 'https://graph.microsoft.com/v1.0/me/MailFolders/{folder_id}/childfolders',
		},
		'user_folders': {
			'1.0': 'https://outlook.office365.com/api/v1.0/users/{user_id}/Folders',
			'2.0': 'https://graph.microsoft.com/v1.0/users/{user_id}/MailFolders',
		},
		'user_folder': {
			'1.0': 'https://outlook.office365.com/api/v1.0/users/{user_id}/Folders/{folder_id}/messages',
			'2.0': 'https://graph.microsoft.com/v1.0/users/{user_id}/MailFolders/{folder_id}/messages',
		}
	}

	def __init__(self, verify=True):
		""" Creates a new inbox wrapper.

		:param verify: whether or not to verify SSL certificate
		"""
		self.url = FluentInbox._get_url('inbox')
		self.fetched_count = 0
		self._filter = ''
		self._search = ''
		self.verify = verify
		self.messages = []

	def from_folder(self, folder_name, user_id=None):
		""" Configure to use this folder for fetching the mails

		:param folder_name: name of the outlook folder
		:param user_id: user id the folder belongs to (shared mailboxes)
		"""
		self._reset()

		folder_id = self.get_folder(value=folder_name,
									by='DisplayName',
									user_id=user_id)['Id']

		if user_id:
			self.url = FluentInbox._get_url('user_folder').format(
				user_id=user_id, folder_id=folder_id)
		else:
			self.url = FluentInbox._get_url('folder').format(
				folder_id=folder_id)

		return self

	def get_folder(self, value, by='Id', parent_id=None, user_id=None):
		"""
		Return a folder by a given attribute.  If multiple folders exist by
		this attribute, only the first will be returned

		Example:
		   get_folder(by='DisplayName', value='Inbox')

		   or

		   get_folder(by='Id', value='AAKrWFG...')

		   Would both return the requested folder attributes

		:param value: Value that we are searching for
		:param by: Search on this key (default: Id)
		:param user_id: user id the folder belongs to (shared mailboxes)
		:returns: Single folder data
		"""
		if parent_id:
			folders_url = FluentInbox._get_url('child_folders').format(
				folder_id=parent_id)
		elif user_id:
			folders_url = FluentInbox._get_url('user_folders').format(
				user_id=user_id)
		else:
			folders_url = FluentInbox._get_url('folders')

		response = Connection.get_response(folders_url,
										   verify=self.verify,
										   params={'$top': 100})

		folder_id = None
		all_folders = []

		for folder in response:
			if folder[by] == value:
				return(folder)

			all_folders.append(folder['displayName'])

		if not folder_id:
			raise RuntimeError(
				'Folder "{}" is not found by "{}", available folders '
				'are {}'.format(value, by, all_folders))

	def list_folders(self, parent_id=None, user_id=None):
		"""
		:param parent_id: Id of parent folder to list.  Default to top folder
		:return: List of all folder data
		"""
		if parent_id:
			folders_url = FluentInbox._get_url('child_folders').format(
				folder_id=parent_id)
		elif user_id:
			folders_url = FluentInbox._get_url('user_folders').format(
				user_id=user_id)
		else:
			folders_url = FluentInbox._get_url('folders')

		response = Connection.get_response(folders_url,
										   verify=self.verify,
										   params={'$top': 100})

		folders = []
		for folder in response:
			folders.append(folder)

		return folders

	def filter(self, filter_string):
		""" Set the value of a filter. More information on what filters are available can be found here:
		https://msdn.microsoft.com/office/office365/APi/complex-types-for-mail-contacts-calendar#RESTAPIResourcesMessage
		More improvements coming soon

		:param filter_string: The string that represents the filters you want to enact.
				should be something like: (HasAttachments eq true) and (IsRead eq false) or just: IsRead eq false
				test your filter string here: https://outlook.office365.com/api/v1.0/me/messages?$filter=
				if that accepts it then you know it works.
		"""
		self._filter = filter_string
		return self

	def search(self, search_string):
		""" Set the value of a search. More information on what searches are available can be found here:
		https://msdn.microsoft.com/office/office365/APi/complex-types-for-mail-contacts-calendar#RESTAPIResourcesMessage
		More improvements coming soon

		:param search_string: The search string you want to use

		Should be something like: "Category:Action AND Subject:Test" or just: "Subject:Test".

		Test your search string here: "https://outlook.office365.com/api/v1.0/me/messages?$search="
		or directly in your mailbox, if that accepts it then you know it works.
		"""
		self._search = search_string
		return self

	def fetch_first(self, count=10):
		""" Fetch the first n messages, where n is the specified count

		:param count: no.of messages to fetch
		"""
		self.fetched_count = 0
		return self.fetch_next(count=count)

	def skip(self, count):
		""" Skips the first n messages, where n is the specified count

		:param count: no.of messages to skip
		"""
		self.fetched_count = count
		return self

	def fetch(self, count=10):
		""" Fetch n messages from the result, where n is the specified count

		:param count: no.of messages to fetch
		"""
		return self.fetch_next(count=count)

	def fetch_next(self, count=1):
		""" Fetch the next n messages after the previous fetch, where n is the specified count

		:param count: no.of messages to fetch
		"""
		skip_count = self.fetched_count
		if self._search:
			params = {'$filter': self._filter, '$top': count,
					  '$search': '"{}"'.format(self._search)}
		else:
			params = {'$filter': self._filter, '$top': count,
					  '$skip': skip_count}

		response = Connection.get_response(self.url, verify=self.verify,
										   params=params)
		self.fetched_count += count

		messages = []
		for message in response:
			messages.append(Message(message, Connection().auth))

		return messages

	@staticmethod
	def _get_url(key):
		""" Fetches the url for specified key as per the connection version configured

		:param key: the key for which url is required
		:return: URL to use for requests
		"""
		return FluentInbox.url_dict[key][Connection().api_version]

	def _reset(self):
		""" Resets the current reference """
		self.fetched_count = 0
		self.messages = []



import logging
import os
import os.path as path
import json

import requests
from oauthlib.oauth2 import TokenExpiredError
from requests_oauthlib import OAuth2Session
from future.utils import with_metaclass


log = logging.getLogger(__name__)


class MicroDict(dict):
	def __getitem__(self, key):
		result = super(MicroDict, self).get(key[:1].lower() + key[1:], None)
		if result is None:
			result = super(MicroDict, self).get(key[:1].upper() + key[1:])
		return result


class Singleton(type):
	_instance = None

	def __call__(cls, *args, **kwargs):
		if not cls._instance:
			cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
		return cls._instance

	# def __new__(cls, *args, **kwargs):
	#	 if not cls._instance:
	#		 cls._instance = object.__new__(cls)
	#	 return cls._instance


_default_token_file = '.o365_token'
_home_path = path.expanduser("~")
default_token_path = path.join(_home_path, _default_token_file)


def save_token(token, token_path=None):
	""" Save the specified token dictionary to a specified file path

	:param token: token dictionary returned by the oauth token request
	:param token_path: path to where the files is to be saved
	"""
	if not token_path:
		token_path = default_token_path

	with open(token_path, 'w') as token_file:
		json.dump(token, token_file, indent=True)


def load_token(token_path=None):
	""" Save the specified token dictionary to a specified file path

	:param token_path: path to the file with token information saved
	"""
	if not token_path:
		token_path = default_token_path

	token = None
	if path.exists(token_path):
		with open(token_path, 'r') as token_file:
			token = json.load(token_file)
	return token


def delete_token(token_path=None):
	""" Save the specified token dictionary to a specified file path

	:param token_path: path to where the token is saved
	"""
	if not token_path:
		token_path = default_token_path

	if path.exists(token_path):
		os.unlink(token_path)


class Connection(with_metaclass(Singleton)):
	_oauth2_authorize_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/authorize'
	_oauth2_token_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/token'

	def __init__(self):
		""" Creates a O365 connection object """
		self.api_version = None
		self.auth = None

		self.oauth = None
		self.client_id = None
		self.client_secret = None
		self.token = None

		self.proxy_dict = None

	def is_valid(self):
		valid = False

		if self.api_version == '1.0':
			valid = True if self.auth else False
		elif self.api_version == '2.0':
			valid = True if self.oauth else False

		return valid

	@staticmethod
	def login(username, password):
		""" Connect to office 365 using specified username and password

		:param username: username to login with
		:param password: password for authentication
		"""
		connection = Connection()

		connection.api_version = '1.0'
		connection.auth = (username, password)
		return connection

	@staticmethod
	def oauth2(client_id, client_secret, store_token=True, token_path=None):
		""" Connect to office 365 using specified Open Authentication protocol

		:param client_id: application_id generated by https://apps.dev.microsoft.com when you register your app
		:param client_secret: secret password key generated for your application
		:param store_token: whether or not to store the token in file system, so u don't have to keep opening
			the auth link and authenticating every time
		:param token_path: full path to where the token should be saved to
		"""
		connection = Connection()

		connection.api_version = '2.0'
		connection.client_id = client_id
		connection.client_secret = client_secret

		if not store_token:
			delete_token(token_path)

		token = load_token(token_path)

		if not token:
			connection.oauth = OAuth2Session(client_id=client_id,
											 redirect_uri='https://outlook.office365.com/owa/',
											 scope=['https://graph.microsoft.com/Mail.ReadWrite',
													'https://graph.microsoft.com/Mail.Send',
													'offline_access'], )
			oauth = connection.oauth
			auth_url, state = oauth.authorization_url(
				url=Connection._oauth2_authorize_url,
				access_type='offline')
			print('Please open {} and authorize the application'.format(auth_url))
			auth_resp = input('Enter the full result url: ')
			os.environ['OAUTHLIB_RELAX_TOKEN_SCOPE'] = 'Y'
			token = oauth.fetch_token(token_url=Connection._oauth2_token_url,
									  authorization_response=auth_resp, client_secret=client_secret)
			save_token(token, token_path)
		else:
			connection.oauth = OAuth2Session(client_id=client_id,
											 token=token)

		return connection

	@staticmethod
	def proxy(url, port, username, password):
		""" Connect to Office 365 though the specified proxy

		:param url: url of the proxy server
		:param port: port to connect to proxy server
		:param username: username for authentication in the proxy server
		:param password: password for the specified username
		"""
		connection = Connection()

		connection.proxy_dict = {
			"http": "http://{}:{}@{}:{}".format(username, password, url, port),
			"https": "https://{}:{}@{}:{}".format(username, password, url,
												  port),
		}
		return connection

	@staticmethod
	def get_response(request_url, **kwargs):
		""" Fetches the response for specified url and arguments, adding the auth and proxy information to the url

		:param request_url: url to request
		:param kwargs: any keyword arguments to pass to the requests api
		:return: response object
		"""
		connection = Connection()

		if not connection.is_valid():
			raise RuntimeError('Connection is not configured, please use "O365.Connection" '
							   'to set username and password or OAuth2 authentication')

		con_params = {}
		if connection.proxy_dict:
			con_params['proxies'] = connection.proxy_dict
		con_params.update(kwargs)

		log.info('Requesting URL: {}'.format(request_url))

		if connection.api_version == '1.0':
			con_params['auth'] = connection.auth
			response = requests.get(request_url, **con_params)
		else:
			try:
				response = connection.oauth.get(request_url, **con_params)
			except TokenExpiredError:
				log.info('Token is expired, fetching a new token')
				token = connection.oauth.refresh_token(Connection._oauth2_token_url, client_id=connection.client_id,
													   client_secret=connection.client_secret)
				log.info('New token fetched')
				save_token(token)

				response = connection.oauth.get(request_url, **con_params)

		log.info('Received response from URL {}'.format(response.url))

		if response.status_code == 401:
			raise RuntimeError('API returned status code 401 Unauthorized, check the connection credentials')

		response_json = response.json()
		if 'value' not in response_json:
			raise RuntimeError('Something went wrong, received an unexpected result \n{}'.format(response_json))

		response_values = [MicroDict(x) for x in response_json['value']]
		return response_values



from O365.message import Message
import logging
import json
import requests

log = logging.getLogger(__name__)

class Inbox( object ):
	'''
	Wrapper class for an inbox which mostly holds a list of messages.
	
	Methods:
		getMessages -- downloads messages to local memory.
		
	Variables: 
		inbox_url -- url used for fetching emails.
	'''
	#url for fetching emails. Takes a flag for whether they are read or not.
	inbox_url = 'https://outlook.office365.com/api/v1.0/me/messages'

	def __init__(self, auth, getNow=True, verify=True):
		'''
		Creates a new inbox wrapper. Send email and password for authentication.
		
		set getNow to false if you don't want to immedeatly download new messages.
		'''
		
		log.debug('creating inbox for the email %s',auth[0])
		self.auth = auth
		self.messages = []
		self.errors = ''

		self.filters = ''
		self.order_by = ''
		self.verify = verify
		
		if getNow:
			self.filters = 'IsRead eq false'
			self.getMessages()

	def getMessages(self, number = 10):
		'''
		Downloads messages to local memory.
		
		You create an inbox to be the container class for messages, this method
		then pulls those messages down to the local disk. This is called in the
		init method, so it's kind of pointless for you. Unless you think new
		messages have come in.

		You can filter only certain emails by setting filters. See the set and
		get filters methods for more information.

				Returns true if there are messages. Returns false if there were no 
				messages available that matched the filters specified. 
		'''

		log.debug('fetching messages.')			
		response = requests.get(self.inbox_url,auth=self.auth,params={'$orderby':self.order_by, '$filter':self.filters, '$top':number},verify=self.verify)
		if response.status_code in [400, 500]:
			self.errors = response.text
			return False
		elif response.status_code in [401]:
			self.errors = response.reason
			return False

		log.info('Response from O365: %s', str(response))

		#check that there are messages
		try:
			response.json()['value']
		except KeyError as e:
			log.debug('no messages')
			return False

		for message in response.json()['value']:
			try:
				duplicate = False
				for i,m in enumerate(self.messages):
					if message['Id'] == m.json['Id']:
						self.messages[i] = Message(message,self.auth)
						duplicate = True
						break
				
				if not duplicate:
					self.messages.append(Message(message,self.auth))

				log.debug('appended message: %s',message['Subject'])
			except Exception as e:
				log.info('failed to append message: %',str(e))

		log.debug('all messages retrieved and put in to the list.')
		return True

	def getErrors(self):
		return self.errors

	def getOrderBy(self):
		return self.order_by

	def setOrderBy(self, f_string):
		'''
		For example 'DateTimeReceived desc'
		'''
		self.order_by = f_string
		return True

	def getFilter(self):
		'''get the value set for a specific filter, if exists, else None'''
		return self.filters

	def setFilter(self,f_string):
		'''
		Set the value of a filter. More information on what filters are available
		can be found here:
		https://msdn.microsoft.com/office/office365/APi/complex-types-for-mail-contacts-calendar#RESTAPIResourcesMessage
		I may in the future have the ability to add these in yourself. but right now that is to complicated.
		
		Arguments:
			f_string -- The string that represents the filters you want to enact.
				should be something like: (HasAttachments eq true) and (IsRead eq false)
				or just: IsRead eq false
				test your filter stirng here: https://outlook.office365.com/api/v1.0/me/messages?$filter=
				if that accepts it then you know it works.
		'''
		self.filters = f_string
		return True

#To the King!
